/*
    YoctOS: LPC2148 Dev Board Operating System

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.        

    (C) Jason Hunt
    nulluser@gmail.com
    
    File: serial.c
*/

#include <core.h>

uart_type uart0;
  
  
/* Setup uart0 
 Dependencies: signal_init, interrupt_init
*/
void serial_init(void)  
{
    // Setup queue

    uart0.tx_head = 0;
    uart0.tx_tail = 0;
    uart0.rx_head = 0;
    uart0.rx_tail = 0;
    uart0.tx_bytes= 0;
    uart0.sending = 0;

//    uart0.tx_lock = signal_create(1);

   	PINSEL0 |= 0x00000005;                

	U0LCR = 0x83;     // 8N1

    // Setup Divisor
	U0DLM = ((PCLK / 16) / uart0_baud) / 256;
	U0DLL = ((PCLK / 16) / uart0_baud) % 256;

	U0LCR = 0x03;    // DLAB = 0      

    // Clear interrupt id, rx register and line status
    U0IIR;  U0RBR;  U0LSR; 

    U0FCR = 0x07;
    U0IER = 0x02;   // Enable data avail interrupt, tx ready interrupt

    // Connected to VIC slot
    if (interrupt_hook(UART0_INT, (void *)serial_uart0_isr))
        panic("Unable to hook serial\n");
}
/* End of serial_init */

       

/* Get char from serial port */
byte serial_rx(byte *ch)
{
	while(!(U0LSR & 0x01));

	*ch = U0RBR;

    return(OK);
}
/* End of serial_rx */



/* Write char to serial port */
void serial_tx(byte ch)
{
    // If interrupts are disabled we need to send and poll until complete
    if (get_cpsr() & IRQ_MASK)
    {
        U0THR = ch;
        while (!(U0LSR & 0x20));
    }
    else
    {
        // Busy wait until a slot is free
        while (uart0.tx_bytes >= TX_BUFFER_SIZE);

        // Protect queue from the tx isr
        enter_critical();

        // If tx buffer is empty, just sent it
        if (uart0.sending == 0)
        {
            U0THR = ch;
            uart0.sending = 1;
        }
        else // Add to queue
        {
            uart0.tx_buffer[uart0.tx_head++] = ch;
            uart0.tx_head %= TX_BUFFER_SIZE;
            uart0.tx_bytes ++;
        }

        leave_critical();
    }

}
/* End of serial_tx */



/* Write char to serial port */
void serial_uart0_isr(void)
{
    save_context();

    dword int_id;                           // Interrupt ID 

    // Process all interrupts
    while ((int_id = (U0IIR & UART_IIR_MASK )) != UART_NO_INRT)
    {
        switch(int_id)
        {
            // The tx buffer has a free slot
            case UART_TX_READY:
            {
                if (uart0.tx_bytes > 0)
                {
                    U0THR = uart0.tx_buffer[uart0.tx_tail++];
                    uart0.tx_tail %= TX_BUFFER_SIZE;
                    uart0.tx_bytes--;
                } 
                else 
                {
                    U0THR;                  // Clear interrupt
                    uart0.sending = 0;      // No longer sending
                } 
            } break;

            // RX line status error
            case UART_RLS: U0LSR; break;
            
            // RX data is ready
            case UART_RX_READY: // rx data
            {
                if (uart0.rx_bytes == TX_BUFFER_SIZE)
                    panic("RX overflow");

                // Add byte to rx buffer
                uart0.rx_buffer[uart0.tx_head++] = U0RBR;
                uart0.rx_head %= RX_BUFFER_SIZE;
                uart0.rx_bytes++;

            } break;
                
            // RX Timeout
            case UART_TIMEOUT:   U0RBR;  break;     // Time out
        
            default:
            {     
                panic("Unknown intntype: %x", int_id);
            
                U0THR;   U0LSR;  U0RBR;     // Clear all sources
            }
        }
    }

    VICVectAddr = ACK_INTERRUPT;             // ACK
    
    restore_context();
}
/* End of serial_tx */







/* Return number of pending tx bytes in buffer */
dword serial_tx_bytes( void )
{
    return(uart0.tx_bytes);
    
}
/* End serial_tx_bytes */




