/*
    YoctOS: LPC2148 Dev Board Operating System

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.        

    (C) Jason Hunt
    nulluser@gmail.com
    
    File: video.c
*/


#include <core.h>
#include "../include/drivers/font.h"

static byte video_cur_x = 0;           // Current x position
static byte video_cur_y = 0;           // Current y position

static byte video_y_offset = 1;              // Offset for scrolling

static word video_f_color = VIDEO_BLACK;     // Current foregound color
static word video_b_color = VIDEO_WHITE;     // Current background color

static byte video_started = false;


/* Setup SPI port */
static void lcd_spi_init(void)
{
    // Enable SPI0 multeplexing for P0.4  Clock  SCK0, and  P0.6  Data   MOSI0
    PINSEL0 |= 0x00001100; 
    
    // Set reset line and chip select lines as outputs
    IODIR0 |= LCD_RESET | LCD_CS;    
    
    // Make sure LCD is disabled
    IOSET0 |= LCD_CS;

    //  BitEnable = 1
    //  CPHA = 1  CPOL = 1  MSTR = 1  LSFB = 0  SPIE = 0
    // BITS = 1001 for 9 bits

    S0SPCR = 0x0000093c;

    S0SPCCR = 8;    // Number of PCLK for SPI clock
}
/* End of spi_init */

   
/* Send command to LCD module */
static void lcd_command(byte data)
{    
    IOCLR0 |= LCD_CS;               // Select display

    S0SPDR = data;                 // Send command

    while ((S0SPSR & SPIF) == 0);   // Wait for transfer complete

    IOSET0 |= LCD_CS;               // Unselect display
}
/* End of lcd_command */


/* Send data to LCD module */
static void lcd_data(byte data)
{
    IOCLR0 |= LCD_CS;                       // Select display

    S0SPDR = (dword)data | 0x0100;   // Send data

    while ((S0SPSR & SPIF) == 0);   // Wait for transfer complete

    IOSET0 |= LCD_CS;                       // Unselect display
}
/* End of lcd_data */


/* Preform soft reset */
static void lcd_reset(void)
{
     // Reset the module
    IOCLR0 = LCD_RESET;		
    sleep(10);

    IOSET0 = LCD_RESET;		
    sleep(10);   
}
/* End of video reset */


/* Setup the LCD 
 Dependencies: None
*/
void video_init(void)
{
    lcd_spi_init();
    
    lcd_reset();
    
#ifdef PHILLIPS

	lcd_command(SWRES);             // Software reset
    sleep(10);

	lcd_command(SLEEPOUT);         // Disable speel mode
//	lcd_command(INVON);            // Invert display 

//    lcd_command(COLMOD);            // Enable 8 bpp mode
//	lcd_data(0x02);	   

    lcd_command(COLMOD);            // Enable 12 bit mode
	lcd_data(0x03);	   

//	lcd_command(MADCTL);	// default values for MAC
//	lcd_data(0x00); // 0xc8
	
    // Setup scrolling
	lcd_command(SRCENB);
	lcd_data(1);	
	lcd_data(129);	
 	lcd_data(2);	

    lcd_command(SETEP);	    // Set scroll set point
    lcd_data(video_y_offset);	 

    video_set_contrast(LCD_CONTRAST);
#endif

#ifdef EPSON
/*    lcd_command(DISCTL);  	// display control(EPSON)
    lcd_data(0x0C);   	// 12 = 1100 - CL dividing ratio [don't divide] switching period 8H (default)
    lcd_data(0x20);
    lcd_data(0x02);

    lcd_command(COMSCN);  	// common scanning direction(EPSON)
    lcd_data(0x01);

    lcd_command(OSCON);  	// internal oscialltor ON(EPSON)

    lcd_command(SLPOUT);  	// sleep out(EPSON)
	
    lcd_command(PWRCTR); 	// power ctrl(EPSON)
    lcd_data(0x0F);         //everything on, no external reference resistors

    lcd_command(DISINV);    // invert display mode(EPSON)

    lcd_command(DATCTL);  	// data control(EPSON)
    lcd_data(0x03);	//correct for normal sin7
    lcd_data(0x00);   	// normal RGB arrangement
    lcd_data(0x01);   	// 8-bit grayscale  
   
    lcd_command(VOLCTR);  	// electronic volume, this is the contrast/brightness(EPSON)
    lcd_data(0x18);         // volume (contrast) setting - fine tuning, original
    lcd_data(0x03);         // internal resistor ratio - coarse adjustment  3
    */
#endif

    sleep(50);

	lcd_command(DISPON);	
	
	sleep(100);
    video_clear();		
   
    video_started = 1;
}
/* End of lcd init */



/*
    Low Level functions
*/

/* Move to setup a windows and enter write mode */
static void lcd_window(byte x, byte y, byte mx, byte my)
{
    lcd_command(PASET);      // page address
	lcd_data(y);             // Y
	lcd_data(my);            // display_y
  
	lcd_command(CASET) ;     // Column Address
	lcd_data(x);             // X
	lcd_data(mx);            // Display_x
  
	lcd_command(RAMWR);    // Write
}
/* End of lcd_window */


/* Macro for writing one color */
#define write_color(c1) \
{ \
    lcd_data(c1 >> 4); \
    lcd_data((c1 & 0x0f) << 4); \
    lcd_command(NOP); \
} \


/* Macro for writing two consecutive colors to the lcd */
#define write_color_pair(c1, c2) \
{ \
    lcd_data(c1 >> 4); \
    lcd_data(((c1 & 0x0f) << 4) | (c2 >> 8)); \
    lcd_data(c2 & 0xff); \
} \


/* Scroll the LCD display up one line */
static void lcd_scroll(void)
{
    word i = 0;

    lcd_window (0, video_y_offset, VIDEO_X-1, VIDEO_Y-1);

    // Erase what will become the next blank line
    for (i = 0; i < FONT_Y * VIDEO_X / 2; i++)
        write_color_pair(video_b_color, video_b_color);

    lcd_command(NOP);        // Exit write mode

    // Move to next line in offset
    video_y_offset += FONT_Y;

    // Wrap offset
    if (video_y_offset >= TEXT_ROWS) video_y_offset = 1;     

    lcd_command(SETEP);	    // Set scroll set point
    lcd_data(video_y_offset);	         
}
/* End of scroll LCD */


/* Write a char to video memory at a location */
static void lcd_char(byte x, byte y, byte c) 
{
    // Apply scroll position to y         
    y = (y + video_y_offset) % TEXT_ROWS;    
        
    // Setup char window
    lcd_window (x, y, x + FONT_X - 1, y + FONT_Y - 1);

 	byte iy;     // Index y
        
    // Scan each row 
    for (iy = 0; iy < FONT_Y; iy++)
    {
        // Get font mask for this char
        byte font_data = font_map[(FONT_Y * (int)(c - 32)) + iy];
        byte mask = 0x80;
        
        // Rotate mask to select which bit to test
        while(mask > 0x80 >> FONT_X)
        {
            // Get color data for two pixels
            word c1 = font_data & mask ? video_f_color : video_b_color;  mask >>= 1;
            word c2 = font_data & mask ? video_f_color : video_b_color;  mask >>= 1;

            write_color_pair(c1, c2);
        }  
    }

    // Exit out of write mode
    lcd_command(NOP);
}
/* End of lcd_char */



/*
    Interface Funtions
*/


/* Clear screen */
void video_clear(void)
{
    lcd_window (0, 0, VIDEO_X, VIDEO_Y);

    word i = 0;

    while (i++ < (VIDEO_X * VIDEO_Y) >> 1)
        write_color_pair(video_b_color, video_b_color);

    lcd_command(NOP);
}
/* End of video_clear */


/* Set a single pixel. For testing only */
void video_put_pixel(byte x, byte y, word color)
{
    lcd_window (x, y, VIDEO_X, VIDEO_Y);
	write_color(color);
}
/* End of put_pixel */


/* Draw a filled box */
void video_filled_box(byte x, byte y, byte xs, byte ys, word color)
{
    lcd_window (x, y, x + xs - 1, y + ys - 1);

    word i = 0;

    while (i++ < ((xs * ys) >> 1)+1)
        write_color_pair(color, color);

    lcd_command(NOP);
}
/* End of filled_box */


/* Write char to LCD module */
void video_putch(byte ch)
{
    if (!video_started) return;
    
    // Newline advance
    if (ch == 0x0d || ch == 0x0a)
    {
        video_cur_y++;
        video_cur_x = 0;
    } else
    {
        if (ch < 0x20 || ch > 0x7f) ch = '-';

        lcd_char(video_cur_x * FONT_X + 3, video_cur_y * FONT_Y,  ch);

        // Advance char x position
        if (++video_cur_x >= TEXT_X)
        {
            video_cur_x = 0;
            video_cur_y ++;
        }
    }

    // See if we need to scroll from last update
    if (video_cur_y >= TEXT_Y)
    {
        video_cur_y = TEXT_Y - 1;     // Clamp y position
        lcd_scroll();               // Scroll LCD up one line
    }
}
/* End of put ch */


/* Set contrast of display */
void video_set_contrast(byte c)
{ 
	lcd_command(SETCON);           // Set Contrast
	lcd_data(c);	
}
/* End of video_set_contrast */


/* Set backgroung color */
void video_set_background(word c)
{
    video_b_color = c;
}
/* End of video_set_background */


/* Get backgroung color */
word video_get_background(void)
{
    return(video_b_color);    
}
/* End of video_get_background */


/* Set foreground color */
void video_set_foreground(word c)
{
    video_f_color = c;
}
/* End of video_set_foreground */


/* Get backgroung color */
word video_get_foreground(void)
{
    return(video_f_color);    
}
/* End of video_get_background */


/* Set current char positon */
void video_set_pos(byte x, byte y)
{
    video_cur_x = x;
    video_cur_y = y;
}
/* End of video_set_pos */


/* Set current char positon */
void video_get_pos(byte *x, byte *y)
{
    *x = video_cur_x;
    *y = video_cur_y;
}
/* End of video_set_pos */

