/*
    YoctOS: LPC2148 Dev Board Operating System

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.        

    (C) Jason Hunt
    nulluser@gmail.com

    Memory is laid out in blocks followed by the area that is returned to the caller.

    [Size  prev  next  checksum (16 bytes) ] [data]  
    
    The blocks are stacked end to end in a linear fashion 

    File: memory.c
*/


#include <core.h>

// Bounds for memory, constants filled in by linker
static const byte * free_start = (byte *) &heap_start;
static const byte * free_end = (byte *) &heap_end;

static block_type *start_block = NULL;  // Main block


/* Returns checksum for a block */
static dword get_checksum(block_type *b)
{
    return((b->size * (dword)(b)) & ~BLOCK_USED);
}
/* End of get_checksum */


/* Setup the memory system by making the first free block 
  Dependencies: None.
*/
void memory_init(void)
{
    // Setup start_block to point to the start of free memory
    start_block = (block_type *) free_start;

    start_block->prev = NULL;
    start_block->next = NULL;

    // Setup size of free area in the main block
    // Remember to account for the size of the header
    start_block->size = free_end - free_start - sizeof(block_type);

    // Calculate the checksum, used flag clear
    start_block->checksum = get_checksum(start_block);
}
/* End of memory_init */


/* Allocate a block of memory */
void * kmalloc(dword size)
{
    if (!size) return(NULL);

    enter_critical();

    block_type *cur_block = start_block ;   // Start with main block

    // We need to align the size to a dword boundry (4 bytes)
    size += 4 - (size & 0x03);

    // Search the blocks for enough space to allocate
    while(cur_block)
    {
        // The block is free
        if ((cur_block->checksum & BLOCK_USED) == 0)
        {
            // See if block is exactly the same size
            if (cur_block->size == size)
            {
                // This block is now used
                cur_block->checksum |= BLOCK_USED;

                leave_critical();

                // Return the actual storage area
                return((void *)((dword)cur_block + sizeof(block_type))); 
            }

            // See if there is room for the this block,
            // The next block, and a sizeof(block_type) byte padding
            // There is no sense in squeezing a block in there 
            // if it can't hold any data, So bail if MIN_SPLIT not met

            if (cur_block->size > size + sizeof(block_type) + MIN_SPLIT)
            {
                // Point to new free area
                block_type *new_free = (block_type *)((dword)cur_block + size + sizeof(block_type));

                // Setup poiters for new block
                new_free -> prev = cur_block;
                new_free -> next = cur_block->next;                
                new_free->size = cur_block->size - size - sizeof(block_type); // Remaining 
                new_free->checksum = get_checksum(new_free);

                // Point the prev element of cur_blocks next to the new block
                // Unless we are at the end of the list
                if (cur_block ->next) cur_block->next->prev = new_free;
                
                // Correct old block. Prev is unchanged
                cur_block->next = new_free;
                cur_block->size = size;
                cur_block->checksum = get_checksum(cur_block) | BLOCK_USED;

                leave_critical();
                
                // Return the actual storage area
                return((void *)((dword)cur_block + sizeof(block_type))); 
            }
        }

        cur_block = cur_block->next;
    }

    leave_critical();

    return(NULL);
}
/* End of kmalloc */



/* Free a block of memory */
void kfree(void * addr)
{
    enter_critical();

    // Point to start of block
    block_type *cur_block = (block_type *)((dword)addr - sizeof(block_type) ); 

    // Check block checksum
    if ((cur_block->checksum & ~BLOCK_USED) != get_checksum(cur_block))
    {
        panic("kfree: Block (%x) checksum error, heap corrupt\n Actual %x Calc %x\n",
                cur_block, cur_block->checksum & ~BLOCK_USED, 
                get_checksum(cur_block));
    }

    // Fill the freed block with garbage
    byte *cur_fill = (byte *)cur_block + sizeof(block_type);

    dword i;

    for (i = 0; i < cur_block->size; i++) *cur_fill++ = i;


    // Preform reverse compaction
    if (cur_block->prev  && !(cur_block->prev->checksum & BLOCK_USED))
    {   
        // The previous block will become the new larger block
        cur_block->prev->size = cur_block->size + cur_block->prev->size + sizeof(block_type) ;

        // Calculate the checksum, leave room for used flag
        cur_block->prev->checksum = get_checksum(cur_block);

        // Chain over the previous block
        cur_block->prev->next = cur_block->next;    
        
        // Reset prev pointer for next block, unles we are at the end
        if (cur_block->next) cur_block->next->prev = cur_block->prev;

        // Seek to previous block
        cur_block = cur_block->prev;
    }

    // Preform forward compaction

    // See if next block exists and is free
    if (cur_block->next && !(cur_block->next->checksum & BLOCK_USED))
    {
        // The current block block will become the new larger block
        cur_block->size = cur_block->size + cur_block->next->size + sizeof(block_type);

        // Calculate the checksum, keep used flag clear
        cur_block->checksum = get_checksum(cur_block);
 
        // chain over the block
        cur_block->next = cur_block->next->next;

        // Make sure this is not the last block, and chain over
        if (cur_block->next) cur_block->next->prev = cur_block;    
    }

    cur_block->checksum = get_checksum(cur_block);

    leave_critical();
}
/* End of kfree */


/* Return free memory in bytes */
dword memory_get_free(void)
{
    block_type *cur_block = start_block ;

    if (cur_block == NULL) return(0);
    
    dword free_mem = 0;

    while (cur_block)
    {
        if ((cur_block->checksum & BLOCK_USED) == 0)
            free_mem += cur_block->size;  

        cur_block = cur_block->next;
    }

    return(free_mem);
}
/* End of memory_get_free */



/*
    Debugging 
*/


/* Display the current memory blocks */
#ifdef MEMORY_DEBUG
void show_blocks(void)
{
    block_type *cur_block = start_block ;

    if (cur_block == NULL) return;
    
    dword used_mem = 0, free_mem = 0;

    // Dump memory blocks
    while (cur_block)
    {
        printk("St %x ", cur_block);            // Start
        printk("Sz: %x ", cur_block->size);     // Size
        printk("Nx: %x ", cur_block->next);     // Next
        printk("Pr: %x ", cur_block->prev);     // Previous
        printk("C: %x ", cur_block->checksum);  // Checksum

        if (cur_block->checksum & BLOCK_USED)
        {
            printk("U\n");
            used_mem += cur_block -> size;  
        }
        else
        {
            printk("F\n");
            free_mem += cur_block -> size;  
        }

        cur_block = (block_type *)cur_block->next;
    }
    
    printk("Used memory: %d\n", used_mem);
    printk("Free memory: %d\n", free_mem);
}
#endif
/* End of  show_blocks */


/* Fill a memory area with random data */
#ifdef MEMORY_DEBUG
static void fill_rand(byte * area, word s)
{
    word i = 0;
    for (i = 0; i < s; i++) area[i] = rand() % 0xff;   
}
#endif
/* End of fill_rand */


/* Attempt to break the memory manager */
#ifdef MEMORY_DEBUG
void memory_test(void)
{
    printk("memory test\n");
    printk("Free start: %x\n", (unsigned long)free_start);

    show_blocks();
    
    const unsigned  test_size = 50;

    unsigned int i;
    void *locations[test_size];

    // Allocate random sizes
    for (i = 0; i < test_size; i++)
    {
        word size = rand() % 50 + 50;

        locations[i] = (void *) kmalloc(size); 
        
        fill_rand((byte *)locations[i], size);
    }

    show_blocks();

    dword j = 100000;

    printk("Testing\n");

    // This should fragment it badly
    while (j--)
    {
        // Choose a random location
        unsigned int l = rand() % test_size; 

        kfree(locations[l]);

        word size = rand() % 50 + 50;

        // Reallocate it
        locations[l] = (void *) kmalloc(size);  

        fill_rand((byte *)locations[l], size);

        if (!locations[l]) 
            printk("Unable to allocate\n");
    }

    printk("Free all\n");

    // free all of them ( reverse)
    for (i = test_size  ;i; i--)
        kfree(locations[i-1]);

    //  free all of them (forward)
//    for (i = 0;i < test_size; i++)
//        kfree(locations[i]);

    show_blocks();
    
    printk("Passed\n");
}
#endif
/* End of mem test */















