/*
    YoctOS: LPC2148 Dev Board Operating System

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.        

    (C) Jason Hunt
    nulluser@gmail.com
    
    File: task.c
*/


#include <core.h>

static const byte * core_size = (byte *) &core_stack_size;     // User stack size 
static const byte * core_addr = (byte *) &core_stack_addr;     // User stack address 

static dword num_tasks = 0;                        // Number of tasks
static dword next_tid = 0;                   // Task id of next stack


// Must be global for save and restore context
task_type *cur_task = NULL;                             // Current task ptr


/* Setup the inital process in the process ring 
 Dependencies: memory_init
*/
void task_init( void )
{
    // Get memory for the task data
    cur_task = (task_type *) kmalloc(sizeof(task_type));

    // Get stack memory address
    cur_task->stack = (dword *)((dword)core_addr - (dword)core_size);
    
    cur_task->stack_size = (dword)core_size;
       
    cur_task->tid = next_tid++;

    cur_task->state = TS_READY;             // Ready and not sleeping
	cur_task->sleep = 0;

    cur_task->wait_signal = 0;


    cur_task->next = cur_task;              // Setup inital ring
   
    num_tasks++;
}
/* End of task_init */


/* This will put a value onto the stack of a process */
void stack_push(task_type *p, dword val)
{
    p->sp -= sizeof(val); 

    *(dword*)p->sp = val;
}
/* End of stack_push */


/* Adds a task to the ring */
byte task_create(void *process, dword stack_size, dword param)
{
    enter_critical();    

    // save old pointer to the next process
    task_type *old_next = cur_task->next; 
    task_type *orig_process = cur_task; 

    // get memory for the process table entry
    cur_task->next = (task_type *) kmalloc(sizeof(task_type));

    if (cur_task->next == NULL)
        panic("task_create: Unable to get memory for task struct\n");

    cur_task = cur_task->next; // Now working with new task

    cur_task->next = old_next; // Complete ring insertion

    // now cur_process is the new process that we are adding
    cur_task->stack = (dword *) kmalloc(stack_size);

    if (cur_task->stack == NULL) 
        panic("task_create: Unable to get memory for task stack\n");

    cur_task->stack_size = stack_size;
    cur_task->tid = next_tid++;
   
    cur_task->wait_signal = 0;
	cur_task->state = TS_READY;                 // Ready and not sleeping
	cur_task->sleep = 0;
	
    // Compute inital stack pointer from malloced memory
    cur_task->sp = (dword)cur_task->stack + stack_size;

    dword start_stack = cur_task->sp;

    // Setup return frame for restore context
    stack_push(cur_task, (dword)process+4);     // Return address modifed for ISR
    stack_push(cur_task, INITIAL_R14);          // r14 - link register
    stack_push(cur_task, start_stack);         // r13 - User mode stack pointer
    stack_push(cur_task, INITIAL_R12);          // r12
    stack_push(cur_task, INITIAL_R11);          // r11
    stack_push(cur_task, INITIAL_R10);          // r10
    stack_push(cur_task, INITIAL_R09);          // r09
    stack_push(cur_task, INITIAL_R08);          // r08
    stack_push(cur_task, INITIAL_R07);          // r07
    stack_push(cur_task, INITIAL_R06);          // r06
    stack_push(cur_task, INITIAL_R05);          // r05
    stack_push(cur_task, INITIAL_R04);          // r04
    stack_push(cur_task, INITIAL_R03);          // r03
    stack_push(cur_task, INITIAL_R02);          // r02
    stack_push(cur_task, INITIAL_R01);          // r01
    stack_push(cur_task, param);                // param in r0
    stack_push(cur_task, INITIAL_FLAGS);        // status register
    
    
    // To make sure this process is next in line to get loaded from, and not
    // saved to, we set the current process back to what was it was before we
    // added the new process
    cur_task = orig_process; 
    
    num_tasks++;

    leave_critical();    

    task_reschedule();      // Start the task
    
    return(OK);
}
/* End of task_create */


/* Show all information for a task */
#ifdef TASK_DEBUG
void show_task(char *name, task_type *t)
{
    if (t == NULL) return;
    
    printk("%s address: %x\n", name, t);
    printk("%s sp: %x\n", name, t->sp);
    printk("%s next: %x\n",name,  t->next);
    printk("%s stack: %x\n", name, t->stack);
    printk("%s tid: %x\n", name, t->tid);
    printk("%s state: %x\n",name,  t->state);
    printk("%s sleep: %x\n",name,  t->sleep);

    unsigned int i ;

    dword * s = (dword*)t->sp;

    // Dump stack
    for (i = 0; i < 24; i++)
    {
        printk("%x:  %x\n",s, *s);
        s++;            
    }
}
#endif
/* End of show task */


/* Handle timer ISR */
void task_tick_isr( void ) 
{
    save_context();                 // Save into cur_task 

    ticks++;                        // Tick counter

    task_schedule();                // Reschedule

    T0IR = ACK_TIMER;               //  Clear interrupt flag 

    VICVectAddr = ACK_INTERRUPT;    // ACK 
 
    restore_context();              // Restore the next task
}
/* End of task_tisk_isr */


/* Deal with SWI. Force reschedule */
void task_swi_isr( void ) 
{
    // Correct return address so restore context works
	asm volatile ("add lr, lr, #4");

    save_context();                 // Save into cur_task

    task_schedule();                // Reschedule

    restore_context();              // Restire the next task
}
/* End of timer_0_irq */


/* Core scheduling system */
void task_schedule( void )
{
    if (cur_task == NULL) return;

    if ((dword)cur_task->sp < (dword)cur_task->stack)
    {
        panic("Task: %d  stack overflow SP: %x  Stack %x  size: %x\n",
              cur_task->tid, cur_task->sp, cur_task->stack, cur_task->stack_size);
    }

	task_type *p = cur_task->next;

	// Wakeup any sleeping tasks that are ready
	while(p != cur_task)
	{
		if (p->sleep > 0)
		{
			p->sleep--;
		
			if (p->sleep == 0)
				p->state = TS_READY;
		}

		p = p->next;
	}

    // Start checking with next task
	cur_task = cur_task->next;
	
	// Find next ready task
	while(1)
	{
		if (cur_task->state == TS_READY) break;

		cur_task = cur_task->next;
	}	
   
}
/* End if task_schedule */


/* Force reschedule via SWI */
void task_reschedule( void )
{
   asm volatile ("swi");    // Force SWI to reschedule
}
/* End of task_reschedule */


/* Sleep via scheduler */
void task_sleep(unsigned int t)
{
    enter_critical();    

	cur_task->sleep = t / timer_0_interval;
	cur_task->state = TS_SLEEP;
	
    leave_critical();    
	
	task_reschedule();	
}
/* End of task_sleep */


/* Return number of tasks */
dword task_get_tasks( void )
{
    enter_critical();    

    dword ret = num_tasks;
  
    leave_critical();      
        
    return(ret);
}
/* End of task_get_tasks */


/* Return tid of current task */
dword task_get_tid( void )
{
    enter_critical();    
        
    dword ret = cur_task->tid;
  
    leave_critical();      
        
    return(ret);
}
/* End of task_get_tid */


/* Return the current task */
task_type * task_get_current( void )
{
    return(cur_task);
}




